package UI::All;

###----------------------------------------------------------------###
#  Copyright 2004 - Paul Seamons                                     #
#  Distributed under the Perl Artistic License without warranty      #
###----------------------------------------------------------------###

use strict;
use vars qw($VERSION
            $AUTOLOAD
            $INITIALIZED
            $DEBUG
            $DESKTOP
            $OS
            %TK
            @TK
            $TK
            @TK_ARGV
            $IS_TERM
            $TK_FORCE
            );
BEGIN {
  $VERSION = '0.00';
  $DEBUG   = 1;
  $DESKTOP = uc($ENV{'DESKTOP'} || 'NONE');
  $TK{'MSWin32'} = {
    'GUI'  => [qw(Tk Win32GUI Wx SDL)],
    'Term' => [qw(CursesUI)],
  };
  $TK{'Unix'} = {
    'GUI'  => [($DESKTOP eq 'KDE') ? qw(Qt3 Gtk2 Wx Tk SDL) : qw(Gtk2 Qt3 Wx Tk SDL)],
    'Term' => [qw(CursesUI)],
  };
  $OS = $^O || 'Unix';
}

use UI::All::Base;

###----------------------------------------------------------------###

sub new { # not normally used - but here for the using
  my $class = shift || die "Missing class";
  my $self  = ref($_[0]) ? shift : {@_};
  return bless $self, $class;
}

###----------------------------------------------------------------###

### allow for use UI::All qw(Qt3 Gtk2);
sub import {
  my $class = shift;
  push @TK, @_;
}

### allow for --tk Qt3,Gtk2
sub parse_argv {
  my $self = shift;

  if (grep {/^-+\w+$/} @ARGV) {
    require Getopt::Long;
    Getopt::Long::Configure(qw(pass_through no_auto_abbrev));
    Getopt::Long::GetOptions("tk=s"       => \@TK_ARGV,
                             "is_term|nw" => sub { $IS_TERM = 1 },
                             "gui"        => sub { $IS_TERM = 0 },
                             "forcetk"    => \$TK_FORCE,
                             );
    @TK_ARGV = map {split /\s*,\s*/, $_} @TK_ARGV;
  }
}

### Programmer defined - allowable toolkits
### allow for UI::All->tks(qw(Qt3 Gtk2));
### allow for UI::All->tks([qw(Qt3 Gtk2)]);
### allow for UI::All->tks({GUI => [qw(Qt3 Gtk2)]});
### allow for UI::All->tks(qw(Qt3 Gtk2));
### allow for @UI::All::TK = qw(Qt3 Gtk2);
### allow for %UI::All::TK = (OS1 => {GUI => qw(Qt3 Gtk2), Term => qw(CursesUI)}, OS2 => {});
### allow for %UI::All::TK = (GUI => qw(Qt3 Gtk2), Term => qw(CursesUI));
### allow for %UI::All::TK = (OS1 => {GUI => qw(Qt3 Gtk2), Term => qw(CursesUI)});
sub tks {
  my $self = shift;
  if ($#_ != -1) {
    if (! ref $_[0]) {
      @TK = (@_);
    } elsif (UNIVERSAL::isa($_[0], 'HASH')) {
      %TK = %{ shift() };
    } elsif (UNIVERSAL::isa($_[0], 'ARRAY')) {
      @TK = @{ shift() };
    }
    return;
  }

  return \@TK if $#TK != -1;

  my $tks = $TK{$OS} || $TK{'Unix'};
  $tks ||= $self->is_term ? $TK{'Term'} : $TK{'GUI'};
  $tks ||= {};
  return [$tks] if ! ref $tks;
  return $tks   if ref($tks) eq 'ARRAY';

  $tks = $self->is_term ? $tks->{'Term'} : $tks->{'GUI'};
  $tks ||= [];
  return [$tks] if ! ref $tks;
  return $tks   if ref($tks) eq 'ARRAY';

  return [];
}

sub tk {
  my $self = shift;
  if ($#_ != -1) {
    die 'UI::All->tk($tk) may only be set once' if $TK;
    $TK = shift;
  }
  return $TK;
}

sub is_term {
  my $self = shift;
  $IS_TERM = shift if $#_ != -1;
  return $IS_TERM if defined $IS_TERM;
  # TODO try and figure out if we are a terminal or can do GUI
  return undef;
}

###----------------------------------------------------------------###

sub init {
  my $self = shift;
  return if $self->initialized;

  $self->parse_argv;

  if (! $self->tk) {
    my $tks     = $self->tks;   # programmer defined
    my %tks     = map {$_ => 1} @$tks;
    my $is_term = $self->is_term;
    my %tried;
    foreach my $tk (@TK_ARGV, @$tks) {
      next if $tried{$tk};
      $tried{$tk} = 1;

      if (! $tks{$tk} && ! $TK_FORCE) {
        warn "User specified toolkit $tk is not allowed (posibly use --forcetk)(@$tks)\n";
        next;
      }

      if (! UNIVERSAL::can("UI::All::$tk", 'can')
          && ! eval { require "UI/All/$tk.pm" }) {
        # TODO - log sanely
        warn $@;
        next;
      }

      if (defined $is_term) {
        next if $is_term && ! "UI::All::${tk}"->is_term;
        next if ! $is_term && "UI::All::${tk}"->is_term;
      }

      $self->tk($tk);
      last;
    }

    die "Couldn't find a valid tool kit in (@$tks)" if ! $self->tk;
  }

  $self->initialized(1);
}

sub inc { \@INC }

sub initialized {
  my $self = shift;
  $INITIALIZED = shift if $#_ != -1;
  $INITIALIZED;
}

###----------------------------------------------------------------###

sub AUTOLOAD {
  my $self = shift;
  my $args = ref($_[0]) ? shift: {@_};  # args to be passed to the new object

  $self->init;

  ### allow for bootstrapping the parent application

  ### get the method name
  my $ref = ref($self) || $self; # default case $self is a class
  if (index($AUTOLOAD, $ref.'::') != 0) {
    die "Method $AUTOLOAD was called by an object of type $ref, cross calling not supported";
  }
  my $meth = substr($AUTOLOAD, length($ref) + 2);
  if (index($meth, '::') != -1) {
    die "Not sure how $meth still has :: ($meth).  Perhaps the class doesn't inherit UI?";
  }

  ### find that package in the toolkit
  my $pkg = "UI::All::${TK}::${meth}";
  if (! UNIVERSAL::can($pkg, 'can')
      && ! eval { require "UI/All/$TK/$meth.pm" }) {
    die "Unknown or unimplemented method for UI::All::$TK - $meth";
  }
  if (! $pkg->can('main_loop')) { # check if UI::All is allowed to create this type
    die "Only types that are capable of running main_loop may be created by UI::All ($pkg)";
  }

  local $args->{'base_ui_all_app'} = $self;
  return scalar $pkg->new($args);
}

sub DESTROY {}

###----------------------------------------------------------------###

1;

__END__

=head1 NAME

UI::All - Common Perl interface to the multitudinous GUI toolkits.

=head1 SYNOPSIS

  use UI::All;
  my $mw     = UI::All->MainWindow(title => "Run 1");
  my $vbox   = $mw->LayoutVBox;
  my $label  = $vbox->Label( text    => 'Hello, world!');
  my $button = $vbox->Button(text    => 'Quit',
                             clicked => sub { $mw->exit_loop },
                             );
  $mw->main_loop;
  exit;

  # OR #
  my $mw   = UI->MainWindow(title => "Run 2");
  my $vbox = $mw->LayoutVBox;
  $vbox->Label( text    => 'Hello, world (2)!');
  $vbox->Button(text    => 'Quit',
                clicked => [$mw => 'exit_loop'],
                );
  $mw->main_loop;
  exit;

  # OR #
  my $mw = UI->MainWindow(title => "Run 3");
  $mw->Label( text    => 'Hello, world (3)!');
  $mw->Button(text    => 'Quit',
              clicked => [exit_loop => $mw]);
  $mw->main_loop;
  exit;

=head1 DESCRIPTION

Perl UI is a unified interface to many of the different GUI (and non-GUI)
tool kits used for creating User Interfaces.  The goal is that rather
than learning each of the different tool kit interfaces, a developer can
learn the Perl UI interface which will let them code once, and then use
their application on multiple platforms with any of the available toolkits.

Perl UI's API is derived from all of the toolkits.  Much of the widget heirarchy
is based upon Qt.  The calling syntax is close to Tk.

=head1 FAQ

Well this is more of an AQ than an FAQ because I don't know if the Q's are
really all that frequent.  But we'll try and nip them in the bud.

=over 4

=item Don't we already have graphical toolkits?

Yes we do.  There is Gtk, Gtk2, Qt2, Qt3, Win32 GUI, Tk and
Wx perl.  Oh and there are Curses UI, the Term:: package line, SDL, and others.
Each of these work on various platforms.  Each of these have their own API (and
rightly so as each of the graphical toolkits try to follow the API of their
compiled counterparts.

We're not really interested in creating yet another Toolkit (well we sort of are with SDL).
We want to write an application once, and have it appear native under Windows,
under KDE or Gnome, under OSX, and if there are no graphical components (pixmaps or
canvases) we wouldn't mind if it would work in the console.

=item You did this in Perl?  Why not Python, Ruby (your language here) ?

I know perl fairly well.  I don't know python, ruby, (your language here) - or at least
not to an extent that I feel comfortable with.

=item So why not do it in another language?

I would love it if somebody did.

Python has very good support and interface for most of the toolkits that Perl UI covers.
It would be a great candidate for doing the same thing.

=item Why Perl 5 ? Won't Perl 6 make it easier to implement ?

(Ok - this one was a ringer)  We have Perl 5 now.  Perl 6 will be here relatively soon.
I plan on doing the same thing for Perl 6.  However, after Perl 6 comes out, there will
still need to be work to interface the various graphical toolkits again.  The difference
this time is that Perl 6 will be able to interface directly with the core libraries in a
more efficient way than Perl 5 is able to.

A great benefit is that once this is done, then python or ruby on parrot will have access
to the same library.

=item GUI toolkits have huge API's - isn't this a large undertaking?

Well - yes they are and yes this is.  Contributors and members of the project are
certainly welcome and are encouraged to join.

=item Don't you have a long way to go ?

Yes.  Hence the low version number.

=item You don't have (your widget name here) widget done yet.  Where is it ?

Requests are fine and welcome.  We will move along with each widget once
we have a mostly suitable widget in the toolkits that support it.

=back

=head1 CLASSES

Widgets and layouts are created by calling their class name as a method.  Just about
any object type can instantiate any other object type.  This is done so that parent
child relationships can be automatically assumed.  For example:

  my $mw = UI::All->MainWindow; # typically only MainWindows are created by UI::All

  my $vbox = $mw->LayoutVBox;

  my $hbox = $vbox->LayoutHBox;

  my $lab1 = $vbox->Label( text => "Label1" );
  my $lab2 = $hbox->Label( text => "Label2" );
  my $lab3 = $hbox->Label( text => "Label3" );

  $mw->main_loop;

The currently supported classes are as follows (supported signals and properties
are listed - more will be added to the list as soon as they are reasonably supported
in most toolkits):

=over 4

=item MainWindow

The main widget used for display.

=item TODO - finish the list

=cut

=head1 TIMELINE

Approximate order of priority

=over 4

=item Preview Release (Proof of Concept)

    MainWindow
    LayoutVBox
    LayoutHBox
    TextEdit
    LineEdit
    Button
    FileDialog
    ListBox

=item Initial Release

    Table  - Basic
    Canvas - Basic
    ProgressDialog
    MessageDialog
    RadioButton
    CheckBox
    ComboBox
    Slider
    Toolbar
    Menubar
    Actions
    Pixmaps

=item More basic widgets

=item Better Canvas Support

=item More Widgets / Improved widgets

    TreeView
    ListView
    IconView
    ColorBox
    ColorDialog

=item Network Support

=item Better Asyncronous File IO support

=item More standard dialogs

=item System Tray support

=back

=head1 BUGS

* It isn't very easy to subclass a UI::All widget.  This is because
any subclassing really ought to provide functionality in all of
the toolkits.  This isn't really a fixable bug.  I'm sure it will
annoy more than one person.

* It is heavy.  The system uses a double matched heirarchy object
system.  It does it best to be simple - but probably fails.  It
would be better to just right a "light" wrapper around each toolkit.

* Curses::UI is hard to work with.  A more graceful failing toolkit
will be created.

=head1 AUTHOR

Paul Seamons - <perl-ui@seamons.com>

Hopefully Others

=head1 LICENSE

Distributed under the Perl Artistic License without warranty.

Ok it isn't quite that easy.  Perl UI supports the GPL.  Depending
upon the toolkit used then the Artistic can be applied as well.

For example, Qt probably won't play nicely with the Artistic.

We will probably build in an API that allows for express disabling
of various toolkits for license reasons.

=cut
