package UI::All::Base;

###----------------------------------------------------------------###
#  Copyright 2005 - Paul Seamons                                     #
#  Distributed under the Perl Artistic License without warranty      #
###----------------------------------------------------------------###

use strict;
use vars qw($AUTOLOAD
            $WIDGET_N
            @TRACK
            $TRACK
            );
use UI::Signals qw(registered_signals); # purposely don't import attach or detach

BEGIN {
  $WIDGET_N = 0;
  require Data::Dumper;
  *CORE::GLOBAL::dump = sub { print Data::Dumper::Dumper(@_) };
  #$TRACK = 1;
}

END {
  show_track();
}

sub show_track {
  return if ! $TRACK;
  print "---------TRACK--------------\n";
  foreach (@TRACK) {
    print join(", ", @$_)."\n";
  }
  print "---------TRACK--------------\n";
}

###----------------------------------------------------------------###

sub is_term { 0 }

sub new {
  my $class = shift || die;
  my $self  = ref($_[0]) ? {%{$_[0]}} : {@_};
  bless $self, $class;

  if (my $parent = $self->parent) {
    if ($self->is_widget && $parent->is_layout) {
      $self->parent_widget->add_child($self);
    } else {
      $parent->add_child($self); # add to parent's children
    }
  } elsif (! defined wantarray) {
    require Carp;
    &Carp::croak(ref($self) ." created in void context with no parent to inherit from");
  }

  push(@TRACK, ["NEW", ref($self), "$self"]) if $TRACK;

  ### allow remaining subs to know context
  local $self->{'context'} = wantarray ? 'list' : defined(wantarray) ? 'scalar' : 'void';

  $self->init;

  ### set up our toolkit's object proxy
  $self->init_proxy;

  ### allow for any default properties that weren't handled in construction
  foreach my $key (keys %$self) {
    next if $key =~ /^(parent)$/; # a few - don't sets
    if (my $meth = $self->can("set_${key}")) {
      $self->$meth(delete $self->{$key});
    }
  }

  ### allow for installing registered signals
  my $sigs = $self->registered_signals;
  foreach my $key (keys %$sigs) {
    next if ! $self->{$key};
    $self->attach($key, delete $self->{$key});
  }

  return $self;
}

sub init {}

###----------------------------------------------------------------###

sub name : lvalue {
  my $self = shift;
  $self->{'name'} = shift if $#_ != -1;
  $self->{'name'};
}

sub gen_name { (caller) .'-('.(++ $WIDGET_N).')' }

sub proxy : lvalue {
  my $self = shift;
  $self->{'proxy'} = shift if $#_ != -1;
  $self->{'proxy'};
}

sub isa {
  my ($self, $class) = @_;
  $class = "UI::All::${UI::All::TK}::${class}" if index($class, '::') == -1;
  return $self->SUPER::isa($class);
}

sub pack { die "Pack'ing is not supported by UI::All.  Use layouts (LayoutVBox, LayoutHBox)" }

sub attach {
  my $self = shift;
  my $slot = shift; # the slot that the proxy will call - ends up being a redispatched signal

  ### see if we have a signal registered under that name
  ### the proxy object's signal hits our signal and calls it again
  my $signal = $self->registered_signals->{$slot} || do {
    require Carp;
    &Carp::confess("No matching registered sig ($self, $slot)");
  };
  my $proxy = $self->proxy || die "No proxy object installed for $self";

  $self->attach_proxy($slot, $proxy, $signal);
  return $self->UI::Signals::attach($slot, @_);
}

sub detach { shift->UI::Signals::detach(@_) }

sub attach_proxy { die "Stub @_" }
sub detach_proxy { die "Stub @_" }

###----------------------------------------------------------------###

sub parent {
  my $self = shift;
  if ($#_ != -1) {
    require Carp;
    &Carp::croak("Use set_parent");
  }
  $self->{'parent'};
}

sub set_parent {
  my $self = shift;
  my $new  = shift;
  my $old  = $self->parent;
  return if $new == $old;

  ### remove from the old
  $old->remove_child($self);
  $new->add_child($self) if $new;
}

sub add_child {
  my $self = shift;
  push @{ $self->children }, shift;
}

sub remove_child {
  my $self     = shift;
  my $child    = shift || return;
  my $children = $self->children;

  for (reverse (0 .. $#$children)) {
    next if $children->[$_] ne $child;
    splice @$children, $_, 1, ();
  }
}

sub children { shift->{'children'} ||= [] }

###----------------------------------------------------------------###

sub parent_widget {
  my $self   = shift;
  my $parent = $self->parent;
  $parent = $parent->parent while $parent && ! $parent->is_widget;
  $parent;
}

sub parent_layout {
  my $self   = shift;
  my $parent = $self->parent;
  while ($parent && ! $parent->is_layout) {
    if ($parent->is_widget) { # child of a widget before a layout - get that widget's layout
      return $parent->find_or_create_child_layout;
    }
    $parent = $parent->parent;
  }
  return $parent || do { require Carp; Carp::confess("No matching layout for $self") };
}

sub find_or_create_child_layout {
  my $self = shift;
  return $self->{'_default_layout'} ||= do {
    my $layout;
    foreach my $child (@{ $self->children }) {
      next if ! $child->is_layout;
      $layout = $child;
      last;
    }
    if (! $layout) {
      my $type = $self->{'default_layout_type'} || 'LayoutVBox';
      my $name = ($self->name || "$self") .' -- default layout';
      $layout = $self->$type({
        spacing => $self->{'spacing'},
        margin  => $self->{'margin'},
        name    => $name,
      });
    }
    $layout; # return of the do
  };
}

sub is_widget { 1 }

sub is_layout { 0 }

sub is_pixmap { 0 }

###----------------------------------------------------------------###

sub add_property {
  my ($self, $prop, $value) = @_;
  my $callpkg = caller;
  no strict 'refs';
  my $name = $callpkg.'::'.$prop;
  *$name = sub : lvalue {
    my $self = shift;
    $self->{$prop} = shift if $#_ != -1;
    $self->{$prop};
  } if ! defined &$name;
  $self->$prop($value) if $value;
}

sub dump {
  my $self = shift;
  my $ref = ref $self;
  my ($pkg, $file, $line) = caller;
  my $str = "Dump $ref at $file line $line\n";
  require Data::Dumper;
  $str .= &Data::Dumper::Dumper($self);
  return $str;
}

sub exit_loop { die "This object cannot exit the main loop" }

###----------------------------------------------------------------###

sub add {
  if (! $UI::All::TK) {
    require Carp;
    &Carp::croak("\$UI::All::TK has not been set - first object must be called as UI::All->ObjType");
  }

  my $self  = shift;
  my $class = shift;
  my $args  = shift || {};

  ### find that package in the toolkit
  my $pkg  = "UI::All::${UI::All::TK}::${class}";
  my $file = "$pkg.pm";
  $file =~ s|::|/|g;
  if (! UNIVERSAL::can($pkg, 'can')
      && ! eval { require $file }) {
    require Carp;
    if ($@ && $@ !~ /^Can\'t locate/i) {
      warn "$@";
    }
    &Carp::croak("Unknown or unimplemented method for UI::All::${UI::All::TK} - $AUTOLOAD ($file)");
  }

  if (! exists $args->{'parent'} && ref $self) { # set parent if not explicitly set
    $args->{'parent'} = $self;
  }

  return $pkg->new($args);
}

###----------------------------------------------------------------###

### allow for ->Class(key => 'val') to call ->add(Class => {key => 'val'});
sub AUTOLOAD {
  if (! $UI::All::TK) {
    require Carp;
    &Carp::croak("\$UI::All::TK has not been set - first object must be called as UI::All->ObjType");
  }

  my $self = shift;
  my $args = defined($_[0]) ? shift() : {};
  if (! ref $args) {
    if ($#_ % 2) {
      require Carp;
      &Carp::croak("Invalid method $AUTOLOAD called");
    }
    $args = {$args, @_};
  };

  ### get the method name
  my $ref = ref($self) || $self; # allow for $obj->ObjType or __PACKAGE__->ObjType
  if (index($AUTOLOAD, $ref.'::') != 0) {
    die "Method $AUTOLOAD was called by an object of type $ref, cross calling not supported";
  }
  my $new_class = substr($AUTOLOAD, length($ref) + 2);
  if (index($new_class, '::') != -1) {
    die "Not sure how $new_class still has :: ($new_class).  Perhaps the class doesn't inherit ".__PACKAGE__."?";
  }

  return $self->add($new_class, $args);
}

sub DESTROY {
  my $self = shift || return;
  return if $self->{'is_destroying'};
  $self->{'is_destroying'} = 1;
  foreach my $child (reverse @{ $self->children }) {
    $child->DESTROY;
    $child->{'proxy'} = undef;
    $child = undef;
  }
  $self->{'proxy'} = undef;
  push(@TRACK, ["DES", ref($self), "$self"]) if $TRACK;
}

###----------------------------------------------------------------###

=head1 NAME

UI::All::Base - UI Tool Kit interface to multitudinous Graphical Tool Kits

=cut

1;
