package UI::All::Qt3;

###----------------------------------------------------------------###
#  Copyright 2004 - Paul Seamons                                     #
#  Distributed under the Perl Artistic License without warranty      #
###----------------------------------------------------------------###

use vars qw(@_EXPORT
            %SIZE_POLICY
            );
use base qw(UI::All::Base);
use Qt qw(3.006);
BEGIN {
  require Qt::slots;
  %SIZE_POLICY = (
                  min        => Qt::SizePolicy::Minimum(),
                  max        => Qt::SizePolicy::Maximum(),
                  fixed      => Qt::SizePolicy::Fixed(),
                  preferred  => Qt::SizePolicy::Preferred(),
                  expand     => Qt::SizePolicy::Expanding(),
                  min_expand => Qt::SizePolicy::MinimumExpanding(),
                  ignore     => Qt::SizePolicy::Ignored(),
                  );
};

###----------------------------------------------------------------###

@_EXPORT = qw(attach_qt3);

sub import {
  my $pkg = shift;
  my $callpkg = caller;
  no strict 'refs';
  my %EXPORT = map {$_ => 1} @_EXPORT;
  foreach my $sub ($#_ != -1 ? @_ : @_EXPORT) {
    die "Sub $sub is not exported by ".__PACKAGE__ if ! $EXPORT{$sub};
    my $name = $callpkg .'::'. $sub;
    *$name = \&$sub;
  }
}

sub attach_qt3 { ### hook up our Qt signal to our Qt proxy object
  my ($obj_from, $signal, $obj_to, $slot) = @_;
  $obj_from->{'uitk'} ||= $obj_to; # UI::All parent object stored in proxy

  ### get the qt signal signature
  my $sig = $signal =~ m/^\w+\(\s*(.*?)\s*\)$/ ? $1 : die "Missing signature on $signal";
  my @sig = split /\s*,\s*/, $sig;
  $slot =~ s/\(.*\)$//;

  my $name = "slot_${slot}";
  if (! defined &$name) {
    no strict 'refs';
    *$name = sub { Qt::this()->{'uitk'}->$slot(@_) };

    my $ref = ref $obj_from;
    $ref =~ s/^\ +//; # perlQt munges with our package name
    Qt::slots::import($ref, $name => [@sig]);
  }

  $obj_from->connect($obj_from, SIGNAL($signal), SLOT("$name($sig)"));
}

sub attach_proxy { ### hook up our Qt proxy object to our UI::All::Qt3 object
  my ($self, $slot, $proxy, $signal) = @_;

  ### "double pass" the object because of Qt3's object system
  $proxy->attach_qt3($proxy, $signal, $self, $slot);
}



###----------------------------------------------------------------###
### Layout base class
###----------------------------------------------------------------###

package UI::All::Qt3::Layout;
use base qw(UI::All::Qt3);

use vars qw(
            %ALIGN
            %VALIGN
            );
BEGIN {
  %ALIGN = (
            auto    => Qt::AlignAuto(),
            left    => Qt::AlignLeft(),
            right   => Qt::AlignRight(),
            center  => Qt::AlignHCenter(),
            justify => Qt::AlignJustify(),
            );
  %VALIGN = (
             top    => Qt::AlignTop(),
             bottom => Qt::AlignBottom(),
             center => Qt::AlignVCenter(),
             );
}

sub is_widget { 0 }
sub is_layout { 1 }

sub add_widget {
  my ($layout, $widget, $args) = @_;
  if ($layout->isa('LayoutGrid')) {
    die "No add_widget yet for grid layouts";
  } else {
    my $align  = $args->{'align'}  || $widget->{'align'}  || 'auto';
    my $valign = $args->{'valign'} || $widget->{'valign'} || 'auto';
    $layout->proxy->addWidget($widget->proxy,
                              $widget->{'stretch'} || 0,
                              ($ALIGN{$align} || 0) | ($VALIGN{$valign} || 0),
                              );
  }
}

sub add_spacer {
  my ($layout, $spacer) = @_;
  $layout->proxy->addItem($spacer->proxy);
}

sub set_margin  { shift->proxy->setMargin( shift) }
sub set_spacing { shift->proxy->setSpacing(shift) }

###----------------------------------------------------------------###

package UI::All::Qt3::LayoutVBox;
use base qw(UI::All::Qt3::Layout);

sub init_proxy {
  my $self = shift;
  if ($self->parent->is_layout) {
    $self->proxy = Qt::VBoxLayout($self->parent->proxy,
                                  defined($self->{'spacing'}) ? $self->{'spacing'} : -1,
                                  $self->name || $self->gen_name);

  } else {
    $self->proxy = Qt::VBoxLayout($self->parent->proxy,
                                  $self->{'margin'} || 0,
                                  defined($self->{'spacing'}) ? $self->{'spacing'} : -1,
                                  $self->name || $self->gen_name);
  }
}

###----------------------------------------------------------------###

package UI::All::Qt3::LayoutHBox;
use base qw(UI::All::Qt3::Layout);

sub init_proxy {
  my $self = shift;
  if ($self->parent->is_layout) {
    $self->proxy = Qt::HBoxLayout($self->parent->proxy,
                                  defined($self->{'spacing'}) ? $self->{'spacing'} : -1,
                                  $self->name || $self->gen_name);

  } else {
    $self->proxy = Qt::HBoxLayout($self->parent->proxy,
                                  $self->{'margin'} || 0,
                                  defined($self->{'spacing'}) ? $self->{'spacing'} : -1,
                                  $self->name || $self->gen_name);
  }
}

###----------------------------------------------------------------###

package UI::All::Qt3::LayoutGrid;
use base qw(UI::All::Qt3::Layout);

###----------------------------------------------------------------###

package UI::All::Qt3::Spacer;
use base qw(UI::All::Qt3);

sub init_proxy {
  my $self = shift;
  my $w  = $self->{'w'} || 0;
  my $h  = $self->{'h'} || 0;
  my $wp = defined($self->{'w_policy'}) ? $self->{'w_policy'} : 'min';
  my $hp = defined($self->{'h_policy'}) ? $self->{'h_policy'} : 'min';
  $wp = $UI::All::Qt3::SIZE_POLICY{$wp} || $UI::TK::Qt3::SIZE_POLICY{'min'};
  $hp = $UI::All::Qt3::SIZE_POLICY{$hp} || $UI::TK::Qt3::SIZE_POLICY{'min'};
  $self->proxy = Qt::SpacerItem($w, $h, $wp, $hp);

  $self->parent_layout->add_spacer($self) if $self->{'context'} eq 'void';
}

sub is_widget { 0 }
sub change_size {
  my $self = shift;
  die "TODO: not - implemented";
}



###----------------------------------------------------------------###
### The following are Widget classes supported by UI::All::Qt3
###----------------------------------------------------------------###

package UI::All::Qt3::MainWindow;
use base qw(UI::All::Qt3);

sub init_proxy {
  my $self = shift;
  my $argv = [];
  $self->add_property(app => Qt::Application($argv));
  $self->proxy = Qt::MainWindow($self->name || $self->gen_name);
}

sub parent { undef } # MainWindows shouldn't have parents

sub title     { shift->proxy->caption }
sub set_title { shift->proxy->setCaption(shift) }

sub main_loop {
  my $self = shift;
  $self->proxy->show;
  $self->app->setMainWidget($self->proxy);
  $self->app->exec;
}

sub exit_loop {
  my $self = shift;
  $self->DESTROY;
  $self->app->quit;
}

###----------------------------------------------------------------###

package UI::All::Qt3::Dialog;
use base qw(UI::All::Qt3);

sub init_proxy {
  my $self = shift;
  $self->proxy = Qt::Dialog($self->parent_widget->proxy, $self->name || $self->gen_name, 1);
}

sub set_title { shift->proxy->setCaption(shift) }

sub start { shift->proxy->exec }

sub close { shift->proxy->accept }

###----------------------------------------------------------------###

package UI::All::Qt3::Label;
use base qw(UI::All::Qt3);

sub init_proxy {
  my $self = shift;
  $self->proxy = Qt::Label('', $self->parent_widget->proxy);
  $self->parent_layout->add_widget($self);
}

sub text     { shift->proxy->text }
sub set_text { shift->proxy->setText(shift) }

###----------------------------------------------------------------###

{
  package Proxy::PushButton;
  use Qt::isa qw(Qt::PushButton UI::All::Qt3);
}

package UI::All::Qt3::Button;
use base qw(UI::All::Qt3);
import Proxy::PushButton;
use UI::Signals signals => {
  clicked => 'released()',
};

sub init_proxy {
  my $self = shift;
  $self->proxy = Proxy::PushButton('', $self->parent_widget->proxy);
  $self->parent_layout->add_widget($self);
}

sub text     { shift->proxy->text }
sub set_text { shift->proxy->setText(shift) }

###----------------------------------------------------------------###

{
  package Proxy::LineEdit;
  use Qt::isa qw(Qt::LineEdit UI::All::Qt3);
}

package UI::All::Qt3::LineEdit;
use base qw(UI::All::Qt3);
import Proxy::LineEdit;
use UI::Signals signals => {
  changed        => 'textChanged(const QString &)',
  return_pressed => 'returnPressed()',
};

sub init_proxy {
  my $self = shift;
  $self->proxy = Proxy::LineEdit('', $self->parent_widget->proxy);
  $self->parent_layout->add_widget($self);
}

sub text     { shift->proxy->text }
sub set_text { shift->proxy->setText(shift) }

###----------------------------------------------------------------###

{
  package Proxy::TextEdit;
  use Qt::isa qw(Qt::TextEdit UI::All::Qt3);
}

package UI::All::Qt3::TextEdit;
use base qw(UI::All::Qt3);
import Proxy::TextEdit;
use UI::Signals signals => {
  changed        => 'textChanged()',
  return_pressed => 'returnPressed()',
};

sub init_proxy {
  my $self = shift;
  $self->proxy = Proxy::TextEdit('', undef, $self->parent_widget->proxy);
  $self->parent_layout->add_widget($self);
}

sub text     { shift->proxy->text }
sub set_text { shift->proxy->setText(shift) }

###----------------------------------------------------------------###

{
  package Proxy::ListBox;
  use Qt::isa qw(Qt::ListBox UI::All::Qt3);
}

package UI::All::Qt3::ListBox;
use base qw(UI::All::Qt3);
import Proxy::ListBox;
use UI::Signals signals => {
  changed => 'selectionChanged()',
};

sub init_proxy {
  my $self = shift;
  $self->proxy = Proxy::ListBox($self->parent_widget->proxy, $self->name || $self->gen_name);
  $self->parent_layout->add_widget($self);
}

sub insert {
  my $self = shift;
  my $insert = shift;
  $self->proxy->insertStrList(ref($insert) ? $insert : [$insert]);
}


sub has_selection {
  my $p = shift->proxy;
  for (0 .. $p->count - 1) {
    return 1 if $p->isSelected($_);
  }
}

sub selection {
  my $p = shift->proxy;
  if (wantarray) {
    return (grep {$p->isSelected($_)} 0 .. $p->count - 1);
  }
  for (0 .. $p->count - 1) {
    return $_ if $p->isSelected($_);
  }
}

sub get {
  my $p   = shift->proxy;
  my @ret = map { eval { $p->text($_) } } @_;
  return wantarray ? @ret : $ret[0];
}

sub delete {
  my ($self, $first, $last) = @_;
  $last = $first if ! defined $last;
  ($last, $first) = ($first, $last) if $first > $last;
  $self->proxy->removeItem($_) for reverse $first .. $last;
}

sub last { shift->proxy->count - 1 }

sub set_selection {
  my $p = shift->proxy;
  my $bool = shift() ? 1 : 0;
  map { eval { $p->setSelected($_, $bool) } } @_;
}

###----------------------------------------------------------------###

1;

__END__

=head1 NAME

UI::All::Qt3 - Qt flavor of UI::All

=cut

