package UI::All::Tk;

###----------------------------------------------------------------###
#  Copyright 2004 - Paul Seamons                                     #
#  Distributed under the Perl Artistic License without warranty      #
###----------------------------------------------------------------###

use strict;
use Tk ();
use base qw(UI::All::Base);

sub attach_proxy {
  my ($self, $slot, $proxy, $signal) = @_;

  if (! ref $signal) { # normal Tk callback
    $proxy->configure($signal => sub { $self->$slot(@_) });
  } elsif (UNIVERSAL::isa($signal, 'ARRAY')) {
    my ($seq, $sub) = @$signal;
    if (UNIVERSAL::isa($sub, "CODE")) {
      my $copy = $sub;
      $sub = sub { &$copy($self, $slot, $proxy, $signal, @_) };
    } else {
      $sub = sub { $self->$slot(@_) };
    }
    $proxy->bind($seq, $sub);
  } else {
    die "Unknown signal type $signal";
  }
}

###----------------------------------------------------------------###

package UI::All::Tk::Layout;
use base qw(UI::All::Tk);

sub is_widget { 0 }
sub is_layout { 1 }

# -side => top bottom left right
# -expand => 1 0
# -fill => x y none both
# -ipadx internal x padding
# -ipady internal y padding
# -padx  external x padding
# -pady  external y padding

sub add_widget {
  my ($self, $widget, $args) = @_;
  if ($self != $widget->parent) {
    die "Tk only allows for packing a widget in the parent that it was added to - this means that the widget should be created with a parent of LayoutVBox rather than MainWindow and so on";
  }
  my $margin = defined($self->{'margin'}) ? delete($self->{'margin'}) : 0;
  my $align  = defined($self->{'align'})  ? delete($self->{'align'}) : $self->default_align;
  $align = $self->default_align if $align ne 'bottom';

  my $fill_x = $args->{'w_policy'} || $widget->{'w_policy'} || 'max';
  my $fill_y = $args->{'h_policy'} || $widget->{'h_policy'} || 'max';
  $fill_x = 'max' if $fill_x ne 'fixed';
  $fill_y = 'max' if $fill_y ne 'fixed';
  my $fill = ($fill_x eq 'max' && $fill_y eq 'max') ? 'both'
    : ($fill_x eq 'max') ? 'x'
    : ($fill_y eq 'max') ? 'y'
    : 'none';
  $widget->proxy->pack(-ipadx  => $margin,
                       -ipady  => $margin,
                       -fill   => $fill,
                       -expand => ($fill eq 'both') ? 1 : 0,
                       -side   => $align,
                       );
}

sub add_spacer {
  die;
}

###----------------------------------------------------------------###

package UI::All::Tk::LayoutVBox;
use base qw(UI::All::Tk::Layout);

sub init_proxy {
  my $self = shift;
  my $margin = defined($self->{'margin'}) ? delete($self->{'margin'}) : 0;
  $self->proxy = $self->parent->proxy->Frame;
  $self->proxy->pack(-ipadx  => $margin,
                     -ipady  => $margin,
                     -fill   => 'both',
                     -expand => 1,
                     );
}
sub default_align { 'top' }

###----------------------------------------------------------------###

package UI::All::Tk::LayoutHBox;
use base qw(UI::All::Tk::Layout);

sub init_proxy {
  my $self = shift;
  my $margin = defined($self->{'margin'}) ? delete($self->{'margin'}) : 0;
  $self->proxy = $self->parent->proxy->Frame;
  $self->proxy->pack(-ipadx  => $margin,
                     -ipady  => $margin,
                     -fill   => 'both',
                     -expand => 1,
                     );
}
sub default_align { 'left' }

###----------------------------------------------------------------###

package UI::All::Tk::LayoutGrid;
use base qw(UI::All::Tk::Layout);

sub init_proxy { die }

###----------------------------------------------------------------###
### The following are Widget classes supported by UI::All::Tk
###----------------------------------------------------------------###

package UI::All::Tk::MainWindow;
use base qw(UI::All::Tk);
use Tk;

sub init_proxy {
  my $self = shift;
  my $argv = [];
  my %args;
  $self->proxy = MainWindow->new;
}

sub parent { undef } # MainWindows shouldn't have parents

sub set_title { shift->proxy->title(shift) }

sub main_loop {
  my $self = shift;

  MainLoop;
}

sub exit_loop {
  my $self = shift;
#  Tk->main_quit;
  $self->DESTROY;
  $self->show_track;
  exit;
}

###----------------------------------------------------------------###

package UI::All::Tk::Dialog;
use base qw(UI::All::Tk);
use Tk;

sub init_proxy {
  my $self = shift;
  $self->proxy = $self->parent_layout->proxy->Toplevel;
}

sub set_title { shift->proxy->title(shift) }

sub start {
  my $self = shift;
  $self->{'wait'} = 1;
  $self->proxy->waitVariable(\$self->{'wait'});
}

sub close {
  my $self = shift;
  $self->{'wait'} = 0;
  $self->proxy->destroy;
}

###----------------------------------------------------------------###

package UI::All::Tk::Label;
use base qw(UI::All::Tk);

sub init_proxy {
  my $self = shift;
  $self->proxy = $self->parent_layout->proxy->Label; # must belong to the layout
  $self->parent_layout->add_widget($self);
}

sub text     { shift->proxy->cget('-text') }
sub set_text { shift->proxy->configure('-text' => shift) }

###----------------------------------------------------------------###

package UI::All::Tk::Button;
use base qw(UI::All::Tk);
use UI::Signals signals => {
  clicked => '-command',
};

sub init_proxy {
  my $self = shift;
  $self->proxy = $self->parent_layout->proxy->Button;
  $self->{'h_policy'} = 'fixed' if ! exists $self->{'h_policy'};
  $self->parent_layout->add_widget($self);
}

sub text     { shift->proxy->cget('-text') }
sub set_text { shift->proxy->configure('-text' => shift) }

###----------------------------------------------------------------###

package UI::All::Tk::LineEdit;
use base qw(UI::All::Tk);
use UI::Signals signals => {
  changed        => ['<KeyPress>', \&__check_changed],
  pasted         => ['<<Paste>>',  sub {&__was_pasted; &__check_changed}],
  return_pressed => ['<KeyPress-Return>'],
};

sub __check_changed {
  my ($self, $slot, $proxy, $signal) = @_;
  my $str = $proxy->get;
  if (length $str != $self->strlen) {
    $self->strlen = length($str);
    $self->changed($str);
  }
}

sub __was_pasted {
  my ($self, $slot, $proxy, $signal) = @_;
  $self->pasted;
}

sub init_proxy {
  my $self = shift;
  $self->proxy = $self->parent_layout->proxy->Entry(-background => '#ffffff');
  $self->{'h_policy'} = 'fixed' if ! exists $self->{'h_policy'};
  $self->parent_layout->add_widget($self);
  $self->add_property(strlen => (defined($self->{'text'}) ? length($self->{'text'}) : 0));
}

sub text     { shift->proxy->get }
sub set_text { shift->proxy->configure('-text' => shift) }

###----------------------------------------------------------------###

package UI::All::Tk::TextEdit;
use base qw(UI::All::Tk);
use UI::Signals signals => {
  changed        => ['<KeyPress>', \&__check_changed],
  pasted         => ['<<Paste>>',  sub {&__was_pasted; &__check_changed}],
  return_pressed => ['<KeyPress-Return>'],
};

sub __check_changed {
  my ($self, $slot, $proxy, $signal) = @_;
  my $str = $proxy->Contents;
  if (length $str != $self->strlen) {
    $self->strlen = length($str);
    $self->changed(); #$str);
  }
}

sub __was_pasted {
  my ($self, $slot, $proxy, $signal) = @_;
  $self->pasted;
}

sub init_proxy {
  my $self = shift;
  $self->proxy = $self->parent_layout->proxy->Scrolled('Text',
                                                       -background => '#ffffff',
                                                       -scrollbars => 'osoe',
                                                       );
  $self->parent_layout->add_widget($self);
  $self->add_property(strlen => (defined($self->{'text'}) ? length($self->{'text'}) : 0));
}

sub text     { shift->proxy->Contents }
sub set_text { shift->proxy->Contents(shift) }

###----------------------------------------------------------------###

package UI::All::Tk::ListBox;
use base qw(UI::All::Tk);
use UI::Signals signals => {
  changed => ['<ButtonRelease-1>', sub { shift->changed }],
};

sub init_proxy {
  my $self = shift;

  $self->proxy = $self->parent_layout->proxy->Scrolled('Listbox',
                                                       -selectmode => "single",
                                                       -exportselection => 0, # allow multiple to be used at once
                                                       #-width => "100", -height => "5",
                                                       -scrollbars => 'osoe',
                                                       );
  $self->parent_layout->add_widget($self);
}

sub insert {
  my $p = shift->proxy;
  my $insert = shift;
  $p->insert('end', ref($insert) ? @$insert : $insert);
}

sub has_selection {
  my @ref = shift->proxy->curselection;
  return $#ref != -1;
}

sub selection {
  my @ref = shift->proxy->curselection;
  return wantarray ? @ref : $ref[0];
}

sub get {
  my $p = shift->proxy;
  my @ret  = map { eval { $p->get($_) } } @_;
  return wantarray ? @ret : $ret[0];
}

sub delete {
  my ($self, $first, $last) = @_;
  $last = $first if ! defined $last;
  ($last, $first) = ($first, $last) if $first > $last;
  $self->proxy->delete($first, $last);
}

sub last { shift->proxy->size - 1 }

sub set_selection {
  my $p    = shift->proxy;
  my $bool = shift;
  map { $bool ? $p->selectionSet($_) : $p->selectionClear($_) } @_;
}

###----------------------------------------------------------------###

1;

__END__

=head1 NAME

UI::All::Tk - Tk flavor of UI::All

=cut

