package UI::Signals;

###----------------------------------------------------------------###
#  Copyright 2004 - Paul Seamons                                     #
#  Distributed under the Perl Artistic License without warranty      #
###----------------------------------------------------------------###

use strict;
use vars qw($VERSION
            @EXPORT @EXPORT_OK
            %REGISTERED %SIGNALS %SIGNALS_R
            $AUTO_REGISTER
            );

BEGIN {
  $VERSION    = '0.01';
  @EXPORT     = qw(attach detach registered_signals);
  @EXPORT_OK  = qw(attach detach registered_signals
                   register_signal register_signals uis_attach uis_detach);
  %REGISTERED = ();
  %SIGNALS    = ();
  %SIGNALS_R  = ();
  #$AUTO_REGISTER = 1 if ! defined $AUTO_REGISTER;
};

###----------------------------------------------------------------###

sub dump_signals {
  require Data::Dumper;
  print Data::Dumper::Dumper(\%SIGNALS);
}

sub register_signals {
  my $callpkg = shift;
  my $ref     = shift;
  if (UNIVERSAL::isa($ref, 'HASH')) {
    &register_signal($callpkg, $_, $ref->{$_}) foreach keys %$ref;
  } elsif (UNIVERSAL::isa($ref, 'ARRAY')) {
    &register_signal($callpkg, parse_sub($_)) foreach @$ref;
  } else {
    require Carp;
    &Carp::croak("register_signals requires either an arrayref, or a hashref ($ref)");
  }
}

sub register_signal {
  my ($callpkg, $name, $signature) = @_;
  $callpkg = ref $callpkg if ref $callpkg;

  ### short validation
  if (! UNIVERSAL::can($callpkg, 'can')) {
    require Carp;
    &Carp::confess("The package \"$callpkg\" does not appear to be a valid package.");
  } elsif ($name !~ /^\w+$/ || $name !~ /^\D/) {
    require Carp;
    &Carp::confess("The signal name \"$name\" for package \"$callpkg\" does not appear to be a valid name.");
  } elsif (UNIVERSAL::isa($signature, 'CODE')) {
    require Carp;
    &Carp::confess("The signature for \"$name\" in package \"$callpkg\" cannot be a code ref.");
  }

  ### store the signature
  if ($REGISTERED{$callpkg}->{$name} && $REGISTERED{$callpkg}->{$name} ne $signature) {
    require Carp;
    &Carp::confess("Signature mismatch: ".$REGISTERED{$callpkg}->{$name});
  }
  $REGISTERED{$callpkg}->{$name} = $signature;

  ### if no signal by that name for that package exists yet
  ### create a storage space and install a dispatch method
  ### we might eventually validate the signature as well
  if (! UNIVERSAL::can($callpkg, $name)) {
    no strict 'refs';
    local $SIG{__WARN__} = sub { require Carp; &Carp::croak }; # catch if they redefined
    *{"${callpkg}::${name}"} = sub {
      my $self = shift;
      return if ! exists $SIGNALS{$self}->{$name};
      &$_(@_) for @{ $SIGNALS{$self}->{$name}->[0] };
    };
  }
}

sub registered_signals {
  my ($callpkg) = @_;
  $callpkg = ref $callpkg if ref $callpkg;
  return $REGISTERED{$callpkg} ||= {};
}

sub parse_sub {
  my $code = shift || return ('', '', undef);
  if (ref $code) {
    if (UNIVERSAL::isa($code, 'CODE')) {
      return ('', # name
              '', # signature
              $code, # subroutine ref
              );
    } else {
      die "Can only have a method name or code ref when attaching";
    }
  }

  my $name = ($code =~ s/^\s*(\w+)\s*//)        ? $1 : '';
  my $sig  = ($code =~ s/^\s*\(([^\)]*)\)\s*//) ? $1 : '';
  $sig = [split /\s*,\s*/, $sig] if $sig && $sig =~ /\D/;

  return ($name, $sig, undef); # name, signature, code ref
}

sub import {
  my $pkg = shift;
  my $callpkg = caller;

  my $export = $#_ != -1 ? \@_ : \@EXPORT;
  my %OK = map {$_ => 1} @EXPORT, @EXPORT_OK;

  for (my $i = 0; $i <= $#$export; $i ++) {
    my $name = $export->[$i];
    if ($name eq 'signals') {
      register_signals($callpkg, $export->[++ $i]);
    } elsif ($OK{$name}) {
      no strict 'refs';
      local $SIG{__WARN__} = sub {require Carp; &Carp::carp};
      *{"${callpkg}::$name"} = \&{"${pkg}::${name}"};
    } else {
      require Carp;
      &Carp::croak("$name is not exported by ".__PACKAGE__);
    }
  }

  return 1;
}

###----------------------------------------------------------------###

sub attach {
  my ($obj_from, $name_from, $obj_to, $name_to, @the_rest) = @_;
  die "Missing object"      if ! ref $obj_from;

  my ($name, $sig) = parse_sub($name_from);
  die "Invalid name to connect to ($name)" if ! $name;
  # eventually - maybe validate the signature too

  # allow for argument reordering - normal way -    $obj1->attach('meth1', $obj2, 'meth2')
  if (ref($obj_to) eq 'ARRAY') {                  # $obj1->attach('meth1', [$obj2, 'meth2'])
    ($obj_to, $name_to, @the_rest) = @$obj_to;
  }
  if (UNIVERSAL::isa($obj_to, 'CODE')) {          # $obj1->attach('meth1', sub { ... })
    unshift(@the_rest, $name_to) if $#_ == 3; # if they passed us $name_to - add it back
    $name_to = $obj_to;
    $obj_to  = '';
  } elsif (defined($obj_to) && ! ref($obj_to)
           && length($obj_to) && ref($name_to)) {  # $obj1->attach('meth1', 'meth2', $obj2)
    ($obj_to, $name_to) = ($name_to, $obj_to);
  }


  ### allow for automatic installation of signal
  if ($AUTO_REGISTER) {
    register_signal($obj_from, $name, $sig);
  } else {
    my $pkg = ref $obj_from;
    if (! UNIVERSAL::can($pkg, $name)) {
      require Carp;
      Carp::confess("Signal $name is not registered for package $pkg");
    }
  }


  ### add a code ref on to our storage
  my $ref = $SIGNALS{$obj_from}->{$name} ||= [];
  my $coderefs = $ref->[0] ||= [];
  my $objnames = $ref->[1] ||= [];
  my ($name2, $sig2, $code2) = parse_sub($name_to);
  if (ref $obj_to) {
    if ($code2) {
      push @$coderefs, sub { $obj_to->$code2(@the_rest, @_) };
    } elsif ($name2) {
      push @$coderefs, sub { $obj_to->$name2(@the_rest, @_) };
    } else {
      require Carp;
      &Carp::croak("Couldn't find a target for signal \"$name\" ($obj_from, $name_to)");
    }
  } elsif ($code2) {
    if ($#the_rest == -1) {
      push @$coderefs, $code2;
    } else {
      push @$coderefs, sub { &$code2(@the_rest, @_) };
    }
  } else {
    require Carp;
    &Carp::croak("Cannot find a valid target for signal \"$name\" ($obj_from, $name_to) (2)");
  }
  push @$objnames, $obj_to ? "$obj_to" : ""; # store the signature - or a defined value
  die "Code ref / object mismatch ($#$coderefs != $#$objnames)" if $#$coderefs != $#$objnames;

  ### set up a back reference
  if ($obj_to) {
    $SIGNALS_R{$obj_to}->{$name}->{$obj_from} = $ref;
  }

  return 1;
}

sub detach {
  my ($obj_from, $name_from, $obj_to) = @_;
  my ($name, $sig) = parse_sub($name_from);

  if ($obj_from && defined $obj_to) { # undefined will fall below and get wildcarded
    die "Ambiguous detach - too much or too little information supplied (add name_from or remove an object)" if ! $name;
    delete $SIGNALS_R{$obj_to}->{$name}->{$obj_from} if $obj_to;
    my $ref1 = $SIGNALS{$obj_from}->{$name}->[0];
    my $ref2 = $SIGNALS{$obj_from}->{$name}->[1];
    for (my $i = $#$ref2; $i >= 0; $i --) { # start at back (last in first out)
      if (! $ref2->[$i] && ! $obj_to # both false
          || $ref2->[$i] && $obj_to && $ref2->[$i] eq "$obj_to") { # equal
        splice(@$ref1, $i, 1, ());
        splice(@$ref2, $i, 1, ());
        return;
      }
    }
  }

  my $hash;
  if ($obj_to) {
    if (! $name and $hash = $SIGNALS_R{$obj_to}) {
      foreach my $name (keys %$hash) {
        detach($obj_from, $name, $obj_to) if $name;
      }
    } elsif (! $obj_from and $hash = $SIGNALS_R{$obj_to}->{$name}) {
      foreach my $obj_from (keys %$hash) {
        detach($obj_from, $name, $obj_to) if $obj_from;
      }
    }

  } elsif ($obj_from) {
    if (! $name and $hash = $SIGNALS{$obj_from}) {
      foreach my $name (keys %$hash) {
        detach($obj_from, $name, $obj_to) if $name;
      }
    } elsif (! $obj_to and my $array = $SIGNALS{$obj_from}->{$name}->[1]) {
      foreach my $obj_to (reverse @$array) {     # must do in reverse order
        detach($obj_from, $name, $obj_to || ''); # be sure to pass defined but false $obj_to
      }
    }

  } else {
    die 'Usage: detach ($obj1, "method1", $obj2)';
  }
}

*uis_attach = \&attach;
*uis_detach = \&detach;

###----------------------------------------------------------------###

=head1 NAME

UI::Signals - Easy dispatch from signals to mutiple methods.

=head1 SYNOPSIS

  #!/usr/bin/perl -w

  use strict;

  ###-----------------------------------------------###

  ### setup some nice classes
  {
    package Foo;
    use UI::Signals qw(attach detach);
    use UI::Signals signals => {add_it => 'number'};
    sub new {
      my $class = shift || die;
      my $n     = shift || 0;
      bless \ $n, $class;
    }
    sub n :lvalue { ${ +shift } }
    sub add { shift->n += shift || 0 }
  }
  {
    package Bar;
    @Bar::ISA = qw(Foo);
  }
  {
    package Baz;
    @Baz::ISA = qw(Bar);
  }

  ###-----------------------------------------------###
  ### create some objects

  my $foo = Foo->new(1);
  my $bar = Bar->new(2);
  my $baz = Baz->new(3);
  printf '$foo->n = %d, $bar->n = %d, $baz->n = %d'."\n",
    $foo->n, $bar->n, $baz->n;
  # prints '$foo->n = 1, $bar->n = 2, $baz->n = 3';


  ###-----------------------------------------------###
  ### finally - setup some connections

  ### attach by method name
  $foo->attach('add_it', $foo, 'add');
    # OR UI::Signals::attach($foo, 'add_it', $foo, 'add');

  ### attach by method coderef
  $foo->attach('add_it', $bar, sub {my $self = shift; $self->n += shift});

  ### attach an anonymous sub
  $foo->attach('add_it', undef, sub {$baz->n += shift}); # code ref

  ### try out our signals
  $foo->add_it(3);

  printf '$foo->n = %d, $bar->n = %d, $baz->n = %d'."\n",
    $foo->n, $bar->n, $baz->n;
  # prints '$foo->n = 4, $bar->n = 5, $baz->n = 6';
  # UI::Signals->dump_signals;


  ###-----------------------------------------------###
  ### now we can detach connections

  ### detach with specific info
  $foo->detach('add_it', $foo);
    # OR UI::Signals::detach($foo, 'add_it, $foo);

  ### detach all connections of bar to 'add_it'
  UI::Signals::detach(undef, 'add_it', $bar);

  ### detach all connections of bar to anything
  UI::Signals::detach(undef, undef, $bar);

  ### detach last anonymous sub from foo add_it
  $foo->detach('add_it', '');    # defined but false

  ### detach all connections from foo add_it
  $foo->detach('add_it', undef); # undefined and false

  ### detach all connections from foo
  $foo->detach;

  ### see if it worked
  $foo->add_it(3);

  printf '$foo->n = %d, $bar->n = %d, $baz->n = %d'."\n",
    $foo->n, $bar->n, $baz->n;
  # still prints '$foo->n = 4, $bar->n = 5, $baz->n = 6';
  # UI::Signals->dump_signals;

  ###-----------------------------------------------###

  exit;

=head1 DESCRIPTION

UI::Signals allows for registering signals which are
installed as methods in the calling package.  Other
objects can then "attach" or "connect" to those signals.
When an object with an installed signal calls that method,
all objects or anonymous code references that have "attached"
to the signal are called.

=head1 FUNCTIONS / METHODS

=over 4

=item import

=item register_signals

=item register_signal

=item attach

=item detach

=item dump_signals

=back

=head1 AUTHOR

Paul Seamons

=cut

1;
